import collections
from mdtcollections import DotDict, Alias, ExclusiveList
import pytest

registered_types = {}


def typedfixture(*types, **kwargs):
    """This is a decorator that lets us associate fixtures with one or more arbitrary types.
    We'll later use this type to determine what tests to run on the result"""

    def fixture_wrapper(func):
        for t in types:
            registered_types.setdefault(t, []).append(func.__name__)
        return pytest.fixture(**kwargs)(func)

    return fixture_wrapper


TESTDICT = collections.OrderedDict((('a', 'b'),
                                    ('c', 3),
                                    ('d', 'e'),
                                    ('a', 1),
                                    (3, 35)))
@typedfixture('pickleable')
def dotdict():
    dd = DotDict(TESTDICT)
    return dd


class ComposedClass(object):
    delegated = Alias('s.lower')

    def __init__(self):
        self.s = 'ABC'


@typedfixture('pickleable')
def class_with_alias():
    return ComposedClass()


@typedfixture('pickleable')
def exclusive_list_by_identity():
    return ExclusiveList(range(10))


@typedfixture('pickleable')
def exclusive_list_by_len():
    return ExclusiveList(['a', 'bb'], key=len)


pickleable = registered_types['pickleable']




class Alias(object):
    """
    Descriptor that delegates to a child's attribute or method.
    e.g.
    >>> class A(object):
    >>>     childkeys = Alias('child.keys')
    >>>     child = dict()
    >>>
    >>> a = A()
    >>> a.child['key'] = 'value'
    >>> a.childkeys() # calls a.child.keys(), returns ['key']
    ['key']
    """
    def __init__(self, objattr):
        objname, attrname = objattr.split('.')
        self.objname = objname
        self.attrname = attrname

    def __get__(self, instance, owner):
        if instance is None:
            assert owner is not None
            return _unbound_getter(self.objname, self.attrname)
        else:
            proxied = getattr(instance, self.objname)
            return getattr(proxied, self.attrname)

    def __set__(self, instance, value):
        if instance is None:
            raise NotImplementedError()
        else:
            proxied = getattr(instance, self.objname)
            setattr(proxied, self.attrname, value)


def _unbound_getter(objname, methodname):
    def _method_getter(s, *args, **kwargs):
        obj = getattr(s, objname)
        meth = getattr(obj, methodname)
        return meth(*args, **kwargs)
    return _method_getter


class IndexView(object):
    def __init__(self, attr, index):
        self.attr = attr
        self.index = index

    def __get__(self, instance, owner):
        return getattr(instance, self.attr)[self.index]


class Synonym(object):
    """ An attribute (class or intance) that is just a synonym for another.
    """
    def __init__(self, name):
        self.name = name

    def __get__(self, instance, owner):
        return getattr(instance, self.name)

    def __set__(self, instance, value):
        setattr(instance, self.name, value)


class Attribute(object):
    """For overriding a property in a superclass - turns the attribute back
    into a normal instance attribute"""
    def __init__(self, name):
        self.name = name

    def __get__(self, instance, cls):
        return getattr(instance, self.name)

    def __set__(self, instance, value):
        return setattr(instance, self.name, value)



from .descriptors import *
from .dotdict import DotDict
from .exclusive_list import ExclusiveList
from .named_dict import named_dict

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import pytest
from mdtcollections import ExclusiveList

from .fixtures import *


def test_exclusive_list_keys(exclusive_list_by_len):
    ll = exclusive_list_by_len
    assert ll == ['a', 'bb']


    with pytest.raises(KeyError):
        ll.append([1])  # already something w/ length 1

    with pytest.raises(KeyError):
        ll.append('e')  # already something w/ length 1

    with pytest.raises(KeyError):
        ll.append('*d')  # already something w/ length 2

    ll.append(b'eee')

    with pytest.raises(KeyError):
        ll.append('*&(')  # already something w/ length 2

    assert len(ll) == 3


def test_exclusive_list_exclusivity(exclusive_list_by_identity):
    ll = exclusive_list_by_identity
    assert ll == list(range(10))

    with pytest.raises(KeyError):
        ll.append(0)

    with pytest.raises(KeyError):
        ll.insert(5, 0)

    with pytest.raises(KeyError):
        ll.extend([9, 10, 11])

    with pytest.raises(KeyError):
        ll[4] = 0

    with pytest.raises(KeyError):
        ll[4:8] = [0]


def test_list_methods():  # a quick tour of normal list behavior
    ll = ExclusiveList()
    ll.append('a')
    ll.append('b')
    assert ll == ['a', 'b']

    ll.extend([1, 2, 3])  # a b 1 2 3
    assert ll == ['a', 'b', 1, 2, 3]
    ll.insert(0, 'inserted')  # inserted a b 1 2 3
    assert ll == ['inserted', 'a', 'b', 1, 2, 3]

    ll.insert(2, 'middle')  # inserted a middle b 1 2 3
    assert ll == ['inserted', 'a', 'middle', 'b', 1, 2, 3]

    assert ll.pop() == 3  # inserted a middle b 1 2
    assert ll == ['inserted', 'a', 'middle', 'b', 1, 2]
    ll.remove(1)
    assert ll == ['inserted', 'a', 'middle', 'b', 2]

    ll[2:4] = range(5, 10)
    assert ll == ['inserted', 'a', 5, 6, 7, 8, 9, 2]

    # and finally, check that it didn't lose track of any keys
    for item in ll:
        with pytest.raises(KeyError):
            ll.append(item)

    # and that deleted keys were truly deleted
    for item in ['middle', 'b', 1, 3]:
        ll.append(item)



from .dotdict import DotDict

def named_dict(l):
    """ Creates a DotDict from a list of items that have a ``name`` attribute

    Args:
        l (List[object]): list of objects that have a ``name`` or ``__name__`` attribute

    Returns:
        DotDict[str, object]: mapping of objects' names to objects

    Example:
        >>> import moldesign as mdt
        >>> m1 = mdt.from_name('benzene')
        >>> m2 = mdt.from_name('propane')
        >>> d = named_dict([m1, m2, DotDict])
        >>> list(d.keys())
        ['benzene', 'propane', 'DotDict']
        >>> d.propane
        <propane (Molecule), 11 atoms>
        >>> d.DotDict
        moldesign.utils.classes.DotDict
    """
    return DotDict((_namegetter(obj), obj) for obj in l)


def _namegetter(obj):
    try:
        return obj.name
    except AttributeError:
        return obj.__name__



import collections

class ExclusiveList(collections.UserList):
    """ Behaves like a list, but won't allow duplicate items to be added.

    Unlike an OrderedDict, this allows all normal list methods and retains normal their
    time complexity.

    Two items are considered "duplicates" here if they compare equal.
    A "key" callable may be passed to create keys that determine equality.

    Raises:
        KeyError: upon attempts to add duplicate objects
        TypeError: upon attempts to add unhashable objects to the list

    Notes:
        Key functions may be recomputed. If your key function is slow, you should
        add memoization.

    Args:
        iterable (Iterable): items to initialize the list
        key (callable): callable that accepts a single item in the list and returns an
                        object to use for equality comparisons.
    """
    def __init__(self, iterable=None, key=None):
        super(ExclusiveList, self).__init__()
        self._keys = set()
        self._keyfn = self._identity if key is None else key
        if iterable is not None:
            self.extend(iterable)

    def _validate_new_keys(self, iterable):
        keys = set(map(self._keyfn, iterable))
        if len(keys) < len(iterable):
            raise KeyError('Duplicate keys in passed object!')
        collision = keys.intersection(self._keys)
        if collision:
            raise KeyError('Keys already exist: %s' % ','.join(map(str, collision)))
        return keys

    def append(self, obj):
        keys = self._validate_new_keys([obj])
        super(ExclusiveList, self).append(obj)
        self._keys.update(keys)

    def extend(self, iterable):
        items = list(iterable)
        keys = self._validate_new_keys(items)
        super(ExclusiveList, self).extend(items)
        self._keys.update(keys)

    def insert(self, index, obj):
        self[index: index] = [obj]  # just delegate to __setitem__

    def clear(self):
        self._keys.clear()
        super(ExclusiveList, self).clear()

    @staticmethod
    def _identity(obj):
        return obj

    def __setitem__(self, index, objs):
        if not isinstance(index, slice):
            objs = [objs]
            index = slice(index, index+1)
        removekeys = set(map(self._keyfn, self[index]))
        candidate_keys = (self._keys - removekeys)
        newkeys = set(map(self._keyfn, objs))
        collision = candidate_keys.intersection(newkeys)

        if collision:
            raise KeyError('Keys already exist: %s' % ','.join(map(str, collision)))

        super(ExclusiveList, self).__setitem__(index, objs)
        self._keys.clear()
        self._keys.update(newkeys)
        self._keys.update(candidate_keys)

    def remove(self, obj):
        idx = self.index(obj)
        self.pop(idx)

    def pop(self, index=None):
        if index is None:
            index = len(self) - 1

        key = self._keyfn(self[index])
        assert key in self._keys, "Keys and list are inconsistent"
        obj = super(ExclusiveList, self).pop(index)
        self._keys.remove(key)
        return obj



import pytest
from .fixtures import *


def test_dotdict_get(dotdict):
    dd = dotdict
    assert dd == dotdict
    assert dd.get('c', None) == 3
    assert dd.get('ccc', None) is None
    assert dd.a == TESTDICT['a']
    assert dd.d == TESTDICT['d']


def test_dotdict_iterators(dotdict):
    dd = dotdict
    assert len(dd) == len(TESTDICT)
    assert set(dd.keys()) == set(TESTDICT.keys())
    assert set(dd.values()) == set(TESTDICT.values())
    assert dd

    for item in TESTDICT:
        assert item in dd


def test_dotdict_copy(dotdict):
    dd = dotdict.copy()
    assert dd == dotdict
    assert dd._od == dotdict._od

    dd.a = 4
    assert dd.a == dd['a'] == 4
    assert dotdict.a == dotdict['a'] == 1

    del dd.a
    assert 'a' not in dd
    assert 'a' in dotdict


@pytest.mark.screening
def test_dotdict_removals(dotdict):
    dd = dotdict.copy()
    assert dd.pop('_a-a-a', None) is None
    assert dd.pop('d', 4) == 'e'
    assert 'd' not in dd

    dd.d = 5
    assert 'd' in dd

    # item deletion
    del dd.d
    assert 'd' not in dd
    assert len(dd) == len(TESTDICT) - 1
    del dd[3]
    assert 3 not in dd


def test_dotdict_clear(dotdict):
    dd = dotdict.copy()
    dd.clear()
    assert len(dd) == 0
    assert not dd
    assert 'd' not in dd
    assert 'c' not in dd
    assert 'a' not in dd
    with pytest.raises(AttributeError):
        dd.c
    with pytest.raises(KeyError):
        dd['c']


def test_dotdict_introspection(dotdict):
    dd = dotdict.copy()
    assert not hasattr(dd, 'abcd')
    assert hasattr(dd, 'c')
    assert getattr(dd, 'c', None) is 3
    assert getattr(dd, 'ddd', None) is None
    with pytest.raises(AttributeError):
        dd.ddd
    with pytest.raises(KeyError):
        dd['ddd']
    dd['d'] = 'e'


def test_dotdict_consistency(dotdict):
    dd = dotdict.copy()
    dd['k'] = 12345
    assert getattr(dd, 'k') == 12345
    setattr(dd, 'newkey', -42)
    assert dd['newkey'] == -42


def test_dotdict_preserves_ordering(dotdict):
    assert list(dotdict.keys()) == list(TESTDICT.keys())
    assert list(dotdict.values()) == list(TESTDICT.values())
    assert list(dotdict.items()) == list(TESTDICT.items())



# Copyright 2016-2018 Autodesk Inc., 2018 project contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import find_packages, setup
import versioneer

PACKAGE_NAME = 'mdtcollections'

CLASSIFIERS = """\
Development Status :: 4 - Beta
Intended Audience :: Developers
License :: OSI Approved :: Apache Software License
Programming Language :: Python :: 3
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""


with open('requirements.txt', 'r') as reqfile:
    requirements = [x.strip() for x in reqfile if x.strip()]


cmdclass = versioneer.get_cmdclass()

setup(
    name=PACKAGE_NAME,
    version=versioneer.get_version(),
    classifiers=CLASSIFIERS.splitlines(),
    packages=find_packages(),
    install_requires=[],
    url='http://github.com/avirshup/mdtcollections',
    cmdclass=cmdclass,
    license='Apache 2.0',
    author='Aaron Virshup',
    author_email='avirshup+pypi@gmail.com',
    description='Useful collection classes and utilities')



import pytest
import pickle

from .fixtures import *


@pytest.mark.parametrize('objkey', pickleable)
def test_pickling(objkey, request):
    obj = request.getfixturevalue(objkey)
    for iprotocol in (0,1,2):
        x = pickle.dumps(obj, protocol=iprotocol)
        y = pickle.loads(x)
        assert type(y) == type(obj)



import collections

from .descriptors import Alias

class DotDict(object):
    """ An attribute-accessible dictionary that preserves insertion order
    """
    def __init__(self, *args, **kwargs):
        self._od = collections.OrderedDict(*args, **kwargs)
        self._init = True

    def __delattr__(self, item):
        if not self.__dict__.get('_init', False):
            super().__delattr__(item)
        else:
            try:
                del self._od[item]
            except KeyError:
                raise AttributeError()

    def __delitem__(self, key):
        if not self.__dict__.get('_init', False):
            raise TypeError()
        else:
            del self._od[key]

    def __dir__(self):
        return list(self.keys()) + super().__dir__()

    def __getstate__(self):
        return {'od': self._od}

    def __setstate__(self, state):
        self._od = state['od']
        self._init = True

    def copy(self):
        return self.__class__(self._od.copy())

    def copydict(self):
        """ Returns a copy of the core dictionary in its native class
        """
        return self._od.copy()

    def __eq__(self, other):
        try:
            return self._od == other._od
        except AttributeError:
            return False

    def __repr__(self):
        return str(self._od).replace('OrderedDict', self.__class__.__name__)

    def __getattr__(self, key):
        if not self.__dict__.get('_init', False):
            return self.__getattribute__(key)
        if key in self._od:
            return self._od[key]
        else:
            raise AttributeError(key)

    def __setattr__(self, key, val):
        if not self.__dict__.get('_init', False):
            super().__setattr__(key, val)
        else:
            self._od[key] = val

    def __bool__(self):
        return bool(self._od)

    __nonzero__ = __bool__


for _v in ('keys values items __iter__ __getitem__  __len__ __contains__ clear '
           ' __setitem__ pop setdefault get update').split():
    setattr(DotDict, _v, Alias('_od.%s' % _v))